/*
 * Copyright (c) 2013 J. Lewis Muir <jlmuir@imca-cat.org>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package imcacat.procrun;

import java.nio.ByteBuffer;

/**
 * I am a base class for consumers that read data from an
 * {@code InputStream} into a buffer.
 * <p>
 * I check my interrupted status when I can, but I will usually not see that
 * I have been interrupted since I will almost always be blocked reading my
 * {@code InputStream} which is not an interruptible channel. So, the best
 * way to get me to die is to close the {@code InputStream} I am reading.
 */
public abstract class BufferInputStreamConsumer extends InputStreamConsumer {
  protected byte[] buffer;
  protected int bufferPosition;

  /**
   * Creates an instance with the specified buffer size. Calls
   * {@code super()}.
   *
   * @param bufferSize size of the buffer into which I read data from my
   *          input stream
   */
  public BufferInputStreamConsumer(int bufferSize) {
    super();
    initialize(bufferSize);
  }

  /**
   * Creates an instance with the specified buffer size and thread name.
   * Calls {@code super(name)}.
   *
   * @param bufferSize size of the buffer into which I read data from my
   *          input stream
   * @param name thread name
   */
  public BufferInputStreamConsumer(int bufferSize, String name) {
    super(name);
    initialize(bufferSize);
  }

  /**
   * Creates an instance with the specified buffer size and thread group.
   * Calls {@code super(group)}.
   *
   * @param bufferSize size of the buffer into which I read data from my
   *          input stream
   * @param group thread group
   */
  public BufferInputStreamConsumer(int bufferSize, ThreadGroup group) {
    super(group);
    initialize(bufferSize);
  }

  /**
   * Creates an instance with the specified buffer size, thread group, and
   * thread name. Calls {@code super(group, name)}.
   *
   * @param bufferSize size of the buffer into which I read data from my
   *          input stream
   * @param group thread group
   * @param name thread name
   */
  public BufferInputStreamConsumer(int bufferSize, ThreadGroup group, String name) {
    super(group, name);
    initialize(bufferSize);
  }

  /**
   * Creates an instance with the specified buffer size, thread group,
   * thread name, and thread stack size. Calls
   * {@code super(group, name, stackSize)}.
   *
   * @param bufferSize size of the buffer into which I read data from my
   *          input stream
   * @param group thread group
   * @param name thread name
   * @param stackSize thread stack size
   */
  public BufferInputStreamConsumer(int bufferSize, ThreadGroup group, String name, long stackSize) {
    super(group, name, stackSize);
    initialize(bufferSize);
  }

  /**
   * Initializes me with the specified buffer size. Called from my
   * constructors.
   *
   * @param bufferSize size of the buffer into which I read data from my
   *          input stream
   */
  protected void initialize(int bufferSize) {
    buffer = new byte[bufferSize];
    bufferPosition = 0;
  }

  @Override
  public ByteBuffer data() {
    return ByteBuffer.wrap(buffer, 0, bufferPosition);
  }
}