/*
 * Copyright (c) 2013 J. Lewis Muir <jlmuir@imca-cat.org>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package imcacat.procrun;

import java.io.IOException;
import java.nio.ByteBuffer;

/**
 * I read an {@code InputStream}, discarding what I read as I go, until I
 * encounter the end of the stream or an exception is thrown.
 * <p>
 * I check my interrupted status when I can, but I will usually not see that
 * I have been interrupted since I will almost always be blocked reading my
 * {@code InputStream} which is not an interruptible channel. So, the best
 * way to get me to die is to close the {@code InputStream} I am reading.
 */
public class NullInputStreamConsumer extends InputStreamConsumer {
  private int bufferSize;
  private byte[] buffer;

  /**
   * Creates an instance with the specified buffer size. Calls
   * {@code super()}.
   *
   * @param bufferSize size of read buffer used if {@code skip(long)} is not
   *          supported by my input stream
   */
  public NullInputStreamConsumer(int bufferSize) {
    super();
    initialize(bufferSize);
  }

  /**
   * Creates an instance with the specified buffer size and thread name.
   * Calls {@code super(name)}.
   *
   * @param bufferSize size of read buffer used if {@code skip(long)} is not
   *          supported by my input stream
   * @param name thread name
   */
  public NullInputStreamConsumer(int bufferSize, String name) {
    super(name);
    initialize(bufferSize);
  }

  /**
   * Creates an instance with the specified buffer size and thread group.
   * Calls {@code super(group)}.
   *
   * @param bufferSize size of read buffer used if {@code skip(long)} is not
   *          supported by my input stream
   * @param group thread group
   */
  public NullInputStreamConsumer(int bufferSize, ThreadGroup group) {
    super(group);
    initialize(bufferSize);
  }

  /**
   * Creates an instance with the specified buffer size, thread group, and
   * thread name. Calls {@code super(group, name)}.
   *
   * @param bufferSize size of read buffer used if {@code skip(long)} is not
   *          supported by my input stream
   * @param group thread group
   * @param name thread name
   */
  public NullInputStreamConsumer(int bufferSize, ThreadGroup group, String name) {
    super(group, name);
    initialize(bufferSize);
  }

  /**
   * Creates an instance with the specified buffer size, thread group,
   * thread name, and thread stack size. Calls
   * {@code super(group, name, stackSize)}.
   *
   * @param bufferSize size of read buffer used if {@code skip(long)} is not
   *          supported by my input stream
   * @param group thread group
   * @param name thread name
   * @param stackSize thread stack size
   */
  public NullInputStreamConsumer(int bufferSize, ThreadGroup group, String name, long stackSize) {
    super(group, name, stackSize);
    initialize(bufferSize);
  }

  /**
   * Initializes me with the specified buffer size. Called from my
   * constructors.
   *
   * @param bufferSize size of read buffer used if {@code skip(long)} is not
   *          supported by my input stream
   */
  protected void initialize(int bufferSize) {
    this.bufferSize = bufferSize;
    this.buffer = null;
  }

  @Override
  protected int read() throws IOException {
    if (buffer == null) {
      try {
        int numBytesSkipped = (int)stream.skip(Integer.MAX_VALUE);
        if (numBytesSkipped > 0) return numBytesSkipped;
        /*
         * The documentation for the skip method only says the number of
         * skipped bytes is returned. It says nothing about returning -1 to
         * indicate the end of the stream has been reached, so perform an
         * additional read to determine whether we are at the end of the
         * stream.
         */
        int b = stream.read();
        return (b == -1) ? -1 : 1;
      } catch (IOException e) {
        /*
         * Assume skip is not supported by the stream. Switch to using reads
         * into a buffer instead.
         */
        buffer = new byte[bufferSize];
      }
    }
    return stream.read(buffer);
  }

  @Override
  public ByteBuffer data() {
    return ByteBuffer.allocate(0);
  }
}
