/*
 * Copyright (c) 2013 J. Lewis Muir <jlmuir@imca-cat.org>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package imcacat.procrun;

import java.io.InputStream;
import java.nio.ByteBuffer;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.TimeoutException;

/**
 * I run a {@code Process} and consume data written to its native process'
 * standard output and standard error streams.
 * <p>
 * The native process must correctly die and not leave child processes
 * running attached to the standard output and standard error streams.
 * <p>
 * This can be problematic, especially for shells. For example, the command
 * <tt>sh -c 'sleep 5'</tt> would not be good because some <tt>sh</tt>
 * implementations would start a second native process: <tt>sleep</tt>. If
 * the {@code Process} is destroyed (i.e. killed), only the <tt>sh</tt>
 * native process would be killed; the <tt>sleep</tt> native process would
 * remain running. This will make {@code ProcessRunner} unable to correctly
 * respond to being interrupted because the standard output and standard
 * error streams did not get closed when it destroyed the {@code Process}
 * (which killed the <tt>sh</tt> native process, but not the <tt>sleep</tt>
 * native process). Hence the standard output and standard error streams
 * will remain open since the <tt>sleep</tt> native process is attached to
 * them, and the standard output and standard error consumers will remain
 * blocked reading their streams and not die.
 * <p>
 * Better would be <tt>sh -c 'exec sleep 5'</tt> so that there is only one
 * native process (<tt>sleep</tt> which replaced <tt>sh</tt>), and when
 * {@code Process} is destroyed, <tt>sleep</tt> will be killed and its
 * standard output and standard error streams will be closed, and the
 * standard output and standard error consumers will die.
 * <p>
 * An alternative approach would be to ensure the child process detaches
 * from the standard streams of its parent; this way the child process will
 * continue to run, but the standard streams of its parent will be closed
 * when the {@code Process} is destroyed, and the standard output and
 * standard error consumers will die.
 * <p>
 * <strong>Examples</strong>
 *
 * <p>
 * Here's how to run the <tt>uptime</tt> program by creating and starting an
 * instance of me:
 *
 * <pre>
 * ProcessBuilder builder = new ProcessBuilder("uptime");
 * ProcessRunner runner = new ProcessRunner(builder);
 * try {
 *   runner.start();
 *   runner.join();
 * } finally {
 *   runner.interrupt();
 * }
 * ProcessRunnerResult result = runner.result();</pre>
 *
 * <p>
 * Here's how to run the same <tt>uptime</tt> program using one of my static
 * convenience methods:
 *
 * <pre>
 * ProcessBuilder builder = new ProcessBuilder("uptime");
 * ProcessRunnerResult result = ProcessRunner.run(builder);</pre>
 *
 * <p>
 * Here's how to run the same <tt>uptime</tt> program reading its standard
 * output into an 80 byte fixed-size buffer:
 *
 * <pre>
 * ProcessBuilder builder = new ProcessBuilder("uptime");
 * InputStreamConsumer consumer = new FixedSizeBufferInputStreamConsumer(80);
 * ProcessRunnerResult result = ProcessRunner.run(builder, consumer);</pre>
 */
public class ProcessRunner extends Thread {
  private ProcessBuilder builder;
  private InputStreamConsumer stdoutConsumer;
  private InputStreamConsumer stderrConsumer;
  private ProcessRunnerResult result;

  /**
   * Creates an instance with the specified builder. Calls {@code super()}.
   *
   * @param builder used to start the process when I run
   */
  public ProcessRunner(ProcessBuilder builder) {
    super();
    initialize(builder);
  }

  /**
   * Creates an instance with the specified builder and thread name. Calls
   * {@code super(name)}.
   *
   * @param builder used to start the process when I run
   * @param name thread name
   */
  public ProcessRunner(ProcessBuilder builder, String name) {
    super(name);
    initialize(builder);
  }

  /**
   * Creates an instance with the specified builder, thread group, and
   * thread name. Calls {@code super(group, name)}.
   *
   * @param builder used to start the process when I run
   * @param group thread group
   * @param name thread name
   */
  public ProcessRunner(ProcessBuilder builder, ThreadGroup group, String name) {
    super(group, name);
    initialize(builder);
  }

  /**
   * Creates an instance with the specified builder, thread group, thread
   * name, and thread stack size. Calls
   * {@code super(group, null, name, stackSize)}.
   *
   * @param builder used to start the process when I run
   * @param group thread group
   * @param name thread name
   * @param stackSize thread stack size
   */
  public ProcessRunner(ProcessBuilder builder, ThreadGroup group, String name, long stackSize) {
    super(group, null, name, stackSize);
    initialize(builder);
  }

  private void initialize(ProcessBuilder builder) {
    this.builder = builder;
    stdoutConsumer = null;
    stderrConsumer = null;
    result = null;
  }

  /**
   * Returns the standard-output stream consumer. If one has not been set,
   * the default consumer will be used. The default consumer will consume
   * all data into an automatically growing memory buffer.
   * <p>
   * It is only safe to invoke this method before I have been started.
   *
   * @return standard-output consumer or {@code null} if the default
   *         consumer is to be used
   */
  public InputStreamConsumer stdoutConsumer() {
    return stdoutConsumer();
  }

  /**
   * Sets the standard-output stream consumer.
   * <p>
   * It is only safe to invoke this method before I have been started.
   *
   * @param newConsumer consumer to use as the standard-output consumer or
   *          {@code null} to use the default consumer
   *
   * @see #stdoutConsumer()
   */
  public void stdoutConsumer(InputStreamConsumer newConsumer) {
    stdoutConsumer = newConsumer;
  }

  /**
   * Returns the standard-error stream consumer. If one has not been set,
   * the default consumer will be used. The default consumer will consume
   * all data into an automatically growing memory buffer.
   * <p>
   * It is only safe to invoke this method before I have been started.
   *
   * @return standard-error consumer or {@code null} if the default consumer
   *         is to be used
   */
  public InputStreamConsumer stderrConsumer() {
    return stderrConsumer();
  }

  /**
   * Sets the standard-error stream consumer.
   * <p>
   * It is only safe to invoke this method before I have been started.
   *
   * @param newConsumer consumer to use as the standard-error consumer or
   *          {@code null} to use a default consumer
   *
   * @see #stderrConsumer()
   */
  public void stderrConsumer(InputStreamConsumer newConsumer) {
    stderrConsumer = newConsumer;
  }

  @Override
  public void run() {
    Thread shutdownHook = null;
    Process process = null;
    InputStream processStdout = null;
    InputStream processStderr = null;
    Integer exitStatus = null;
    ByteBuffer stdout = null;
    ByteBuffer stderr = null;
    Throwable exception = null;

    try {
      shutdownHook = createShutdownHook(this);
      Runtime.getRuntime().addShutdownHook(shutdownHook);
      ArrayBlockingQueue<InputStreamConsumer> finishedChannel =
          new ArrayBlockingQueue<InputStreamConsumer>(2);

      process = builder.start();
      if (stdoutConsumer == null) {
        stdoutConsumer = createDefaultStdoutConsumer();
      }
      processStdout = process.getInputStream();
      stdoutConsumer.stream(processStdout);
      stdoutConsumer.finishedChannel(finishedChannel);
      stdoutConsumer.start();
      if (!builder.redirectErrorStream()) {
        if (stderrConsumer == null) {
          stderrConsumer = createDefaultStderrConsumer();
        }
        processStderr = process.getErrorStream();
        stderrConsumer.stream(processStderr);
        stderrConsumer.finishedChannel(finishedChannel);
        stderrConsumer.start();
      }

      for (int i = 0; i < ((stderrConsumer == null) ? 1 : 2); i++) {
        InputStreamConsumer finishedConsumer = finishedChannel.take();
        if (exception != null) continue;
        if (!finishedConsumer.hasException()) continue;
        exception = finishedConsumer.exception();
        process.destroy();
        if (finishedConsumer == stdoutConsumer) {
          if (stderrConsumer != null) {
            stderrConsumer.interrupt();
            closeSilently(processStderr);
          }
        } else {
          stdoutConsumer.interrupt();
          closeSilently(processStdout);
        }
      }
      stdoutConsumer.join();
      if (stderrConsumer != null) stderrConsumer.join();

      process.waitFor();
      exitStatus = process.exitValue();
      stdout = stdoutConsumer.data();
      if (stderrConsumer != null) stderr = stderrConsumer.data();
    } catch (Exception ex) {
      if (exception == null) exception = ex;
    } finally {
      destroySilently(process);
      interruptSilently(stdoutConsumer);
      interruptSilently(stderrConsumer);
      closeSilently(processStdout);
      closeSilently(processStderr);
      removeShutdownHookSilently(shutdownHook);
    }

    result = new ProcessRunnerResult(exitStatus, exception, stdout, stderr);
  }

  /**
   * Returns the result of running the process.
   * <p>
   * This method can be called safely from another thread only after
   * ensuring all my actions <i>happen-before</i> the action of calling this
   * method from the other thread. This is the Java memory model
   * <i>happens-before</i> relationship. For example, after I have been
   * started, it would be safe to call this method from another thread after
   * it called my {@code join} method.
   *
   * @return result of running the process
   */
  public ProcessRunnerResult result() {
    return result;
  }

  private InputStreamConsumer createDefaultStdoutConsumer() {
    return new GrowingBufferInputStreamConsumer(
        GrowingBufferInputStreamConsumer.DEFAULT_BUFFER_SIZE, getThreadGroup());
  }

  private InputStreamConsumer createDefaultStderrConsumer() {
    return new GrowingBufferInputStreamConsumer(
        GrowingBufferInputStreamConsumer.DEFAULT_BUFFER_SIZE, getThreadGroup());
  }

  /**
   * Creates and runs a {@code ProcessRunner} with the specified builder,
   * and with the default standard-output consumer and standard-error
   * consumer.
   *
   * @param builder used by the {@code ProcessRunner} instance to start the
   *          process
   *
   * @return result of running {@code ProcessRunner} instance
   *
   * @throws InterruptedException if the current thread is interrupted while
   *           waiting for the {@code ProcessRunner} to finish
   */
  public static ProcessRunnerResult run(ProcessBuilder builder) throws InterruptedException {
    return run(builder, null);
  }

  /**
   * Creates and runs a {@code ProcessRunner} with the specified builder and
   * standard-output consumer, and with the default standard-error consumer.
   *
   * @param builder used by the {@code ProcessRunner} instance to start the
   *          process
   * @param stdoutConsumer consumer to use as the standard-output consumer
   *          or {@code null} to use the default consumer
   *
   * @return result of running {@code ProcessRunner} instance
   *
   * @throws InterruptedException if the current thread is interrupted while
   *           waiting for the {@code ProcessRunner} to finish
   */
  public static ProcessRunnerResult run(ProcessBuilder builder, InputStreamConsumer stdoutConsumer)
      throws InterruptedException {
    return run(builder, stdoutConsumer, null);
  }

  /**
   * Creates and runs a {@code ProcessRunner} with the specified builder,
   * standard-output consumer, and standard-error consumer.
   *
   * @param builder used by the {@code ProcessRunner} instance to start the
   *          process
   * @param stdoutConsumer consumer to use as the standard-output consumer
   *          or {@code null} to use the default consumer
   * @param stderrConsumer consumer to use as the standard-error consumer or
   *          {@code null} to use the default consumer
   *
   * @return result of running {@code ProcessRunner} instance
   *
   * @throws InterruptedException if the current thread is interrupted while
   *           waiting for the {@code ProcessRunner} to finish
   */
  public static ProcessRunnerResult run(ProcessBuilder builder, InputStreamConsumer stdoutConsumer,
      InputStreamConsumer stderrConsumer) throws InterruptedException {
    try {
      return run(builder, stdoutConsumer, stderrConsumer, 0L);
    } catch (TimeoutException e) {
      /*
       * NOT REACHED: Should never be reached since used wait-forever
       * timeout.
       */
      throw new RuntimeException(e);
    }
  }

  /**
   * Creates and runs a {@code ProcessRunner} with the specified builder,
   * standard-output consumer, standard-error consumer, and timeout.
   *
   * @param builder used by the {@code ProcessRunner} instance to start the
   *          process
   * @param stdoutConsumer consumer to use as the standard-output consumer
   *          or {@code null} to use the default consumer
   * @param stderrConsumer consumer to use as the standard-error consumer or
   *          {@code null} to use the default consumer
   * @param timeout timeout in milliseconds to wait for
   *          {@code ProcessRunner} to complete; 0 means wait forever
   *
   * @return result of running {@code ProcessRunner} instance
   *
   * @throws InterruptedException if the current thread is interrupted while
   *           waiting for the {@code ProcessRunner} to finish
   * @throws TimeoutException if the {@code ProcessRunner} has not finished
   *           running within the specified timeout
   */
  public static ProcessRunnerResult run(ProcessBuilder builder, InputStreamConsumer stdoutConsumer,
      InputStreamConsumer stderrConsumer, long timeout) throws InterruptedException,
      TimeoutException {
    ProcessRunner pr = new ProcessRunner(builder);
    pr.stdoutConsumer(stdoutConsumer);
    pr.stderrConsumer(stderrConsumer);
    try {
      pr.start();
      pr.join(timeout);
      if (pr.isAlive()) throw new TimeoutException();
    } finally {
      pr.interrupt();
    }
    return pr.result();
  }

  private static Thread createShutdownHook(final Thread thread) {
    Thread result = new Thread() {
      @Override
      public void run() {
        try {
          thread.interrupt();
          thread.join(5000L);
        } catch (Exception e) {
          /* IGNORED */
        }
      }
    };
    return result;
  }

  private static void destroySilently(Process process) {
    if (process == null) return;
    try {
      process.destroy();
    } catch (Exception e) {
      /* IGNORED */
    }
  }

  private static void interruptSilently(Thread thread) {
    if (thread == null) return;
    try {
      thread.interrupt();
    } catch (Exception e) {
      /* IGNORED */
    }
  }

  private static void closeSilently(InputStream stream) {
    if (stream == null) return;
    try {
      stream.close();
    } catch (Exception e) {
      /* IGNORED */
    }
  }

  private static void removeShutdownHookSilently(Thread hook) {
    if (hook == null) return;
    try {
      Runtime.getRuntime().removeShutdownHook(hook);
    } catch (Exception e) {
      /* IGNORED */
    }
  }
}
